<?php
/**
 * CWAFC_Admin_Action Class
 *
 * Handles the admin functionality.
 *
 * @package Custom Woo Addon For Ceretax
 */

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	header( 'Status: 403 Forbidden' );
	header( 'HTTP/1.1 403 Forbidden' );
	exit;
}

if ( ! class_exists( 'CWAFC_Admin_Action' ) ) {

	/**
	 *  The CWAFC_Admin_Action Class
	 */
	class CWAFC_Admin_Action {

		/**
		 * Construct.
		 */
		public function __construct() {

			add_action( 'admin_enqueue_scripts', array( $this, 'action__cwafc_admin_init' ) );
			// Add settings fields to the new tab.
			add_action( 'woocommerce_settings_tabs_cwafc_ceretax', array( $this, 'action__cwafc_settings_tab' ) );
			// Save settings.
			add_action( 'woocommerce_update_options_cwafc_ceretax', array( $this, 'action__cwafc_update_settings' ) );
			// Display custom fields in the "General" tab of the product edit page.
			add_action( 'woocommerce_product_options_general_product_data', array( $this, 'action__cwafc_add_custom_product_fields' ) );
			// Save custom fields when the product is saved.
			add_action( 'woocommerce_process_product_meta', array( $this, 'action__cwafc_save_custom_product_fields' ) );
			// Add custom field to product category.
			add_action( 'product_cat_add_form_fields', array( $this, 'action__cwafc_add_custom_field_to_product_category' ), 10, 2 );
			add_action( 'product_cat_edit_form_fields', array( $this, 'action__cwafc_edit_custom_field_in_product_category' ), 10, 2 );
			// Save custom field to product category.
			add_action( 'created_product_cat', array( $this, 'action__cwafc_save_custom_field_in_product_category' ), 10, 2 );
			add_action( 'edited_product_cat', array( $this, 'action__cwafc_save_custom_field_in_product_category' ), 10, 2 );
			// Add admin order line item column.
			add_action( 'woocommerce_admin_order_item_headers', array( $this, 'action__cwafc_woocommerce_admin_order_item_headers' ), 20 );
			// Populate the new column with custom data.
			add_action( 'woocommerce_admin_order_item_values', array( $this, 'action__cwafc_woocommerce_admin_order_item_values' ), 10, 3 );
			// Add admin order style and script.
			add_action( 'admin_footer', array( $this, 'action__cwafc_admin_footer_function' ), 20 );
		}

		/**
		 * Register admin min js and admin min css.
		 *
		 * @param string $hook get the settings page name.
		 * @return void
		 */
		public function action__cwafc_admin_init( $hook ) {
			if ( 'woocommerce_page_wc-settings' !== $hook ) {
				return;
			}
			wp_enqueue_script( CWAFC_PREFIX . '_admin_js', CWAFC_URL . 'assets/js/admin.js', array( 'jquery-core' ), CWAFC_VERSION, true );
			wp_enqueue_style( CWAFC_PREFIX . '_admin_css', CWAFC_URL . 'assets/css/admin.css', array(), CWAFC_VERSION );
		}

		/**
		 * Woocommerce setting tabs callback
		 *
		 * @return void
		 */
		public function action__cwafc_settings_tab() {
			woocommerce_admin_fields( $this->cwafc_get_settings() );
		}

		/**
		 * Woocommerce setting tabs save fields
		 *
		 * @return mix
		 */
		public function action__cwafc_update_settings() {

			// phpcs:disable WordPress.Security.NonceVerification.Missing -- Nonce verification already handled in WC_Admin_Settings::save()
			$cere_tax_api_key = ( null !== CWAFC_PREFIX && isset( $_POST[ CWAFC_PREFIX . '_cere_tax_api_key' ] ) ) ? sanitize_text_field( wp_unslash( $_POST[ CWAFC_PREFIX . '_cere_tax_api_key' ] ) ) : '';
			$cere_tax_tax_env = ( null !== CWAFC_PREFIX && isset( $_POST[ CWAFC_PREFIX . '_cere_tax_environment' ] ) ) ? sanitize_text_field( wp_unslash( $_POST[ CWAFC_PREFIX . '_cere_tax_environment' ] ) ) : '';
			// phpcs:enable
			$valid_api_key = $this->cwafc_is_valid_cere_tax_api_key( $cere_tax_api_key, $cere_tax_tax_env );

			if ( ! $valid_api_key ) {
				WC_Admin_Settings::add_error( __( 'API key is not valid. Please add valid API key.', 'ceretax' ) );
				return false;
			} else {
				woocommerce_update_options( $this->cwafc_get_settings() );
				WC_Admin_Settings::add_message( __( 'API key validated successfully.', 'ceretax' ) );
			}
		}

		/**
		 * Define the settings fields
		 *
		 * @return $settings
		 */
		public function cwafc_get_settings() {
			$settings = array(
				array(
					'name' => __( 'Connect to CereTax', 'ceretax' ),
					'type' => 'title',
				),

				array(
					'name'     => __( 'API Key', 'ceretax' ),
					'type'     => 'text',
					'desc'     => __( 'Enter your API key for CereTax', 'ceretax' ),
					'id'       => CWAFC_PREFIX . '_cere_tax_api_key',
					'desc_tip' => true,
					'autoload' => false,
				),

				array(
					'name'     => __( 'Environment', 'ceretax' ),
					'type'     => 'select',
					'desc'     => __( 'Select the environment', 'ceretax' ),
					'id'       => CWAFC_PREFIX . '_cere_tax_environment',
					'desc_tip' => true,
					'autoload' => false,
					'options'  => array(
						'cert'       => __( 'Cert', 'ceretax' ),
						'production' => __( 'Production', 'ceretax' ),
					),
				),

				array(
					'name'     => __( 'Profile', 'ceretax' ),
					'type'     => 'text',
					'desc'     => __( 'Enter your profile for CereTax', 'ceretax' ),
					'id'       => CWAFC_PREFIX . '_cere_tax_profile',
					'desc_tip' => true,
					'autoload' => false,
				),

				array(
					'type' => 'sectionend',
					'id'   => 'cere_tax_section_end',
				),

				array(
					'name' => __( 'General Settings', 'ceretax' ),
					'type' => 'title',
				),

				array(
					'name'    => __( 'Enable CereTax', 'ceretax' ),
					'type'    => 'checkbox',
					'id'      => CWAFC_PREFIX . '_cere_tax_enable',
					'class'   => 'cere_tax_checkbox',
					'default' => 'no',
				),

				array(
					'name'    => __( 'Disable CereTax For Refund Orders', 'ceretax' ),
					'type'    => 'checkbox',
					'id'      => CWAFC_PREFIX . '_cere_tax_refund_disable',
					'class'   => 'cere_tax_checkbox',
					'default' => 'no',
				),

				array(
					'name'    => __( 'Post finalized transactions to CereTax', 'ceretax' ),
					'type'    => 'checkbox',
					'id'      => CWAFC_PREFIX . '_cere_tax_post_transactions',
					'class'   => 'cere_tax_checkbox',
					'default' => 'no',
				),

				array(
					'name'    => __( 'Enable Logging', 'ceretax' ),
					'type'    => 'checkbox',
					'id'      => CWAFC_PREFIX . '_cere_tax_enable_logging',
					'class'   => 'cere_tax_checkbox',
					'default' => 'no',
				),

				array(
					'type' => 'sectionend',
					'id'   => 'cere_tax_section_end',
				),

				array(
					'name' => __( 'Address Validation Settings', 'ceretax' ),
					'type' => 'title',
				),

				array(
					'name'    => __( 'Validate customer and vendor addresses', 'ceretax' ),
					'type'    => 'checkbox',
					'id'      => CWAFC_PREFIX . '_cere_tax_validate_customer_addresses',
					'class'   => 'cere_tax_checkbox',
					'default' => 'no',
				),

				array(
					'name'    => __( 'Validate addresses on every transaction', 'ceretax' ),
					'type'    => 'checkbox',
					'id'      => CWAFC_PREFIX . '_cere_tax_validate_on_transaction',
					'class'   => 'cere_tax_checkbox',
					'default' => 'no',
				),

				array(
					'type' => 'sectionend',
					'id'   => 'cere_tax_section_end',
				),

				array(
					'name' => __( 'Default Settings', 'ceretax' ),
					'type' => 'title',
				),

				array(
					'name'     => __( 'Business Type', 'ceretax' ),
					'type'     => 'text',
					'desc'     => __( 'Enter your business type', 'ceretax' ),
					'id'       => CWAFC_PREFIX . '_cere_tax_business_type',
					'desc_tip' => true,
					'autoload' => false,
				),

				array(
					'name'     => __( 'Customer Type', 'ceretax' ),
					'type'     => 'text',
					'desc'     => __( 'Enter your customer type', 'ceretax' ),
					'id'       => CWAFC_PREFIX . '_cere_tax_customer_type',
					'desc_tip' => true,
					'autoload' => false,
				),

				array(
					'name'     => __( 'PS Code', 'ceretax' ),
					'type'     => 'text',
					'desc'     => __( 'Enter your PS code', 'ceretax' ),
					'id'       => CWAFC_PREFIX . '_cere_tax_ps_code',
					'desc_tip' => true,
					'autoload' => false,
				),

				array(
					'name'     => __( 'Tax Name', 'ceretax' ),
					'type'     => 'text',
					'desc'     => __( 'Enter the Tax Name. If you are using a multilingual plugin do not make any changes here.', 'ceretax' ),
					'id'       => CWAFC_PREFIX . '_cere_tax_name',
					'desc_tip' => true,
					'autoload' => false,
					'default'  => 'Tax',
				),

				array(
					'name'     => __( 'Line Item Tax Column Name', 'ceretax' ),
					'type'     => 'text',
					'desc'     => __( 'Enter the admin order line item tax column name. If you are using a multilingual plugin do not make any changes here.', 'ceretax' ),
					'id'       => CWAFC_PREFIX . '_cere_tax_line_item_column_name',
					'desc_tip' => true,
					'autoload' => false,
					'default'  => 'Tax Total',
				),

				array(
					'name'    => __( 'Tax Included', 'ceretax' ),
					'type'    => 'checkbox',
					'id'      => CWAFC_PREFIX . '_cere_tax_tax_included',
					'class'   => 'cere_tax_checkbox',
					'default' => 'no',
				),

				array(
					'type' => 'sectionend',
					'id'   => 'cere_tax_section_end',
				),
			);

			return $settings;
		}

		/**
		 * Check API key entered is valid or not
		 *
		 * @param string $api_key ceratax api key.
		 * @param string $tax_env ceratax api env.
		 * @return boolean
		 */
		public function cwafc_is_valid_cere_tax_api_key( $api_key, $tax_env ) {

			if ( 'production' === $tax_env ) {
				$api_url = 'https://calc.prod.ceretax.net/test';
			} else {
				$api_url = 'https://calc.cert.ceretax.net/test';
			}

			$response = wp_remote_post(
				$api_url,
				array(
					'method'  => 'POST',
					'headers' => array(
						'x-api-key'    => $api_key,
						'Content-Type' => 'application/json',
					),
					'timeout' => 60,
				)
			);

			if ( is_wp_error( $response ) ) {
				return false;
			} else {
				$response_code = wp_remote_retrieve_response_code( $response );
			}
			if ( 200 === $response_code ) {
				return true;
			}
			return false;
		}

		/**
		 * Add custom fields to products
		 *
		 * @return void
		 */
		public function action__cwafc_add_custom_product_fields() {

			woocommerce_wp_text_input(
				array(
					'id'          => CWAFC_PREFIX . '_cere_tax_product_ps_code',
					'label'       => __( 'CereTax PS Code', 'ceretax' ),
					'placeholder' => __( 'Enter PS Code', 'ceretax' ),
					'desc_tip'    => 'true',
					'description' => __( 'PS Codes help identify the appropriate taxes, taxability, and rates that are applicable to a transaction.', 'ceretax' ),
				)
			);
		}

		/**
		 * Save custom fields to products
		 *
		 * @param string $product_id Product ID.
		 * @return void
		 */
		public function action__cwafc_save_custom_product_fields( $product_id ) {

			$product_ps_code_key = CWAFC_PREFIX . '_cere_tax_product_ps_code';
			$product             = wc_get_product( $product_id );
			// phpcs:disable WordPress.Security.NonceVerification.Missing
			$product_ps_code = isset( $_POST[ $product_ps_code_key ] ) ? sanitize_text_field( wp_unslash( $_POST[ $product_ps_code_key ] ) ) : '';
			// phpcs:enable WordPress.Security.NonceVerification.Missing
			$product->update_meta_data( $product_ps_code_key, sanitize_text_field( $product_ps_code ) );
			$product->save();
		}

		/**
		 * Add custom fields to products category
		 *
		 * @return void
		 */
		public function action__cwafc_add_custom_field_to_product_category() {
			?>
				<div class="form-field">
					<label for="custom_field"><?php esc_html_e( 'CereTax PS Code', 'ceretax' ); ?></label>
					<input type="text" name="<?php echo esc_attr( CWAFC_PREFIX . '_cere_tax_product_cat_ps_code' ); ?>" id="cere_tax_product_cat_ps_code" value="">
					<p class="description"><?php esc_html_e( 'Enter product category PS code.', 'ceretax' ); ?></p>
				</div>
			<?php
		}

		/**
		 * Add custom fields in products edit category
		 *
		 * @param string $term ceratax term object.
		 *
		 * @return void
		 */
		public function action__cwafc_edit_custom_field_in_product_category( $term ) {
			$cere_tax_product_cat_ps_code = get_term_meta( $term->term_id, CWAFC_PREFIX . '_cere_tax_product_cat_ps_code', true );
			?>
			<tr class="form-field">
				<th scope="row" valign="top">
					<label for="cere_tax_product_cat_ps_code"><?php esc_html_e( 'CereTax PS Code', 'ceretax' ); ?></label>
				</th>
				<td>
					<input type="text" name="<?php echo esc_attr( CWAFC_PREFIX . '_cere_tax_product_cat_ps_code' ); ?>" id="cere_tax_product_cat_ps_code" value="<?php echo esc_attr( $cere_tax_product_cat_ps_code ) ? esc_attr( $cere_tax_product_cat_ps_code ) : ''; ?>">
					<p class="description"><?php esc_html_e( 'Enter product category PS code.', 'ceretax' ); ?></p>
				</td>
			</tr>
			<?php
		}

		/**
		 * Add custom fields in products edit category
		 *
		 * @param string $term_id ceratax term ID.
		 *
		 * @return void
		 */
		public function action__cwafc_save_custom_field_in_product_category( $term_id ) {
			$term_field_meta_key = CWAFC_PREFIX . '_cere_tax_product_cat_ps_code';
			// phpcs:disable WordPress.Security.NonceVerification.Missing
			if ( isset( $_POST[ $term_field_meta_key ] ) ) {
				update_term_meta( $term_id, $term_field_meta_key, sanitize_text_field( wp_unslash( $_POST[ $term_field_meta_key ] ) ) );
			}
			// phpcs:enable WordPress.Security.NonceVerification.Missing
		}

		/**
		 * Admin footer script & style
		 *
		 * @return void
		 */
		public function action__cwafc_admin_footer_function() {
			$order_id = null;
			if ( ( isset( $_GET['post'] ) && ! empty( $_GET['post'] ) ) || ( isset( $_GET['id'] ) && ! empty( $_GET['id'] ) ) ) {
				if ( isset( $_GET['post'] ) && ! empty( $_GET['post'] ) ) {
					$order_id = $_GET['post'];
				} else if ( isset( $_GET['id'] ) && ! empty( $_GET['id'] ) ) {
					$order_id = $_GET['id'];
				} else {
					$order_id = '';
				}
			}

			$cere_tax_name = get_option( CWAFC_PREFIX . '_cere_tax_name', 'Tax' );
			$ceretax_data  = get_post_meta( $order_id, 'ceretax_data', true );
			$order         = wc_get_order( $order_id );
			if ( ! empty( $order ) ) {
				foreach ( $order->get_fees() as $fee_id => $fee ) {
					if ( $fee->get_name() === $cere_tax_name ) {
						$fee_amount = $fee->get_amount();
						break;
					}
				}
			}
			if ( ! empty( $order_id ) && ! empty( $ceretax_data ) && $fee_amount > 0 ) { ?>
				<style>
					/* Hide the 3rd tr of .wc-order-totals if .wc-used-coupons is found */
					.wc-order-data-row:has(.wc-used-coupons) .wc-order-totals:first-of-type tr:nth-of-type(3) {
						display: none !important;
					}

					/* Hide the 2nd tr of .wc-order-totals if .wc-used-coupons is not found */
					.wc-order-data-row:not(:has(.wc-used-coupons)) .wc-order-totals:first-of-type tr:nth-of-type(2) {
						display: none !important;
					}

				</style>
			<?php } if ( ! empty( $order_id ) && ! empty( $ceretax_data ) ) { ?>
				<style>
					/* Hide Tax column for ceretax order item list */
					.woocommerce_order_items tr th.line_tax,
					.woocommerce_order_items tr td.line_tax { 
						display: none !important; 
					}

				</style>
			<?php
			}

			// Refund order script.
			if ( ! empty( $order_id ) && ! empty( $ceretax_data ) ) { ?>
			<script>
				jQuery(function($) {
					jQuery( '#woocommerce-order-items' ).on( 'change', 'input.refund_order_item_qty', function() {
						var $row               = $( this ).closest( 'tr.item' );
						var qty                = $row.find( 'input.quantity' ).val();
						var refund_qty         = $( this ).val();
						var ceretax_line_total = $( 'input.ceretax_line_total', $row );
						var line_total         = $( 'input.line_total', $row );
						var refund_line_total  = $( 'input.refund_line_total', $row );

						var unit_line_ceretax_total = accounting.unformat( ceretax_line_total.attr( 'data-line-tax' ), woocommerce_admin.mon_decimal_point ) / qty;
						ceretax_line_total.val(
							parseFloat( accounting.formatNumber( unit_line_ceretax_total * refund_qty, 2, '' ) )
							.toString()
							.replace( '.', woocommerce_admin.mon_decimal_point )
						).trigger( 'change' );

						// Totals
						var unit_total = accounting.unformat( line_total.attr( 'data-total' ), woocommerce_admin.mon_decimal_point ) / qty;

						refund_line_total.val(
							parseFloat( accounting.formatNumber( unit_total * refund_qty, woocommerce_admin_meta_boxes.rounding_precision, '' ) )
								.toString()
								.replace( '.', woocommerce_admin.mon_decimal_point )
						).trigger( 'change' );

						$( this ).trigger( 'refund_quantity_changed' );

					});
					$( '#woocommerce-order-items' ).on( 'change', '.refund input.ceretax_line_total', function() {

						var refund_amount     = 0;
						var $items            = $( '.woocommerce_order_items' ).find( 'tr.item, tr.fee, tr.shipping' );
						var round_at_subtotal = 'yes' === woocommerce_admin_meta_boxes.round_at_subtotal;

						// Restrict value based on data-line-tax
						var maxValue   = parseFloat( $( this ).data('line-tax') ) || 0;
						var inputValue = parseFloat( $( this ).val() ) || 0;

						if ( inputValue > maxValue ) {
							alert( 'You cannot refund the line item tax exceeding the allowed limit of ' + maxValue + '.' );
							$( this ).val( maxValue );
						}

						$items.each(function() {
							var $row               = $( this );
							var refund_cost_fields = $row.find( '.refund input:not(.refund_order_item_qty)' );

							refund_cost_fields.each(function( index, el ) {
								var field_amount = accounting.unformat( $( el ).val() || 0, woocommerce_admin.mon_decimal_point );
								refund_amount += parseFloat( round_at_subtotal ?
									field_amount :
									accounting.formatNumber( field_amount, woocommerce_admin_meta_boxes.currency_format_num_decimals, '' ) );
							});
						});

						$( '#refund_amount' ).val( 
							accounting.formatNumber(
								refund_amount,
								woocommerce_admin_meta_boxes.currency_format_num_decimals,
								'',
								woocommerce_admin.mon_decimal_point
						) ).trigger( 'change' );

					});
				});
			</script>
			<?php }

		}

		/**
		 * Add admin order line item column header.
		 *
		 * @param object $order Order object.
		 * @return void
		 */
		public function action__cwafc_woocommerce_admin_order_item_headers( $order ) {
			$ceretax_data_lineitem = get_post_meta( $order->get_id(), 'ceretax_data_lineitem', true );
			if ( ! empty( $ceretax_data_lineitem ) ) {
				$cere_tax_line_item_column_name = get_option( CWAFC_PREFIX . '_cere_tax_line_item_column_name', 'Tax Total' );
				echo '<th class="ceretax_line_heading" width="7%">' . esc_html( $cere_tax_line_item_column_name ) . '</th>';
			}
		}

		/**
		 * Add admin order line item column header.
		 *
		 * @param object $product Product object.
		 * @param object $item Product item object.
		 * @param string $item_id Order object.
		 * @return void
		 */
		public function action__cwafc_woocommerce_admin_order_item_values( $product, $item, $item_id ) {

			$order_id = null;
			if ( isset( $_GET['post'] ) && ! empty( $_GET['post'] ) ) {
				$order_id = $_GET['post'];
			} else if ( isset( $_GET['id'] ) && ! empty( $_GET['id'] ) ) {
				$order_id = $_GET['id'];
			} else {
				$order_id = null;
			}

			$cere_tax_name         = get_option( CWAFC_PREFIX . '_cere_tax_name', 'Tax' );
			$order                 = wc_get_order( $order_id );
			$ceretax_data_lineitem = json_decode( get_post_meta( $order_id, 'ceretax_data_lineitem', true ) );
			$ceretax_data          = json_decode( get_post_meta( $order_id, 'ceretax_data', true ) );

			if ( ! empty( $ceretax_data_lineitem ) && ! empty( $ceretax_data ) && ! empty( $order ) ) {

				if ( ! empty( $product ) ) {

					// $refunded_qty = -1 * $order->get_total_refunded_for_item( $item_id );

					echo '<td class="ceretax_line_value" width="7%">';
					if ( is_object( $ceretax_data_lineitem ) && is_object( $ceretax_data ) ) {
						$ceretax_items_data = $ceretax_data_lineitem->invoice->lineItems;
						$ceretax_data       = $ceretax_data->invoice->lineItems;
						if ( ! empty( $ceretax_items_data ) && ! empty( $ceretax_data ) ) {

							if ( $product->is_type( 'variation' ) ) {
								$product_id = $product->get_parent_id();
							} else {
								$product_id = $product->get_id();
							}

							$filtered   = array_filter( $ceretax_items_data, function ( $filtered_item ) use ( $product_id ) {
								return isset( $filtered_item->itemNumber ) && $filtered_item->itemNumber == $product_id;
							});

							$refunded_filtered = array_filter( $ceretax_data, function ( $refunded_item ) use ( $product_id ) {
								return isset( $refunded_item->itemNumber ) && $refunded_item->itemNumber == $product_id;
							});
							$total_line_tax          = ! empty( $filtered ) ? reset( $filtered )->totalTaxLine : null;
							$total_line_refunded_tax = ! empty( $refunded_filtered ) ? reset( $refunded_filtered )->totalTaxLine : null;
							$ceretax_refunded_line_item_tax = $item->get_meta( '_ceretax_refunded_line_item_tax' );
							echo '<div class="view">';
								echo wc_price( $total_line_tax, array( 'currency' => $order->get_currency() ) );
								if ( is_array( $order->get_refunds() ) && count( $order->get_refunds() ) > 0 && ! empty( $ceretax_refunded_line_item_tax ) ) {
									echo '<small class="refunded">' . wc_price( $ceretax_refunded_line_item_tax, array( 'currency' => $order->get_currency() ) ) . '</small>';
								}
							echo '</div>';

							echo '<div class="refund" style="display: none;">
								<input type="text" name="ceretax_line_total[' . absint( $item_id ) . ']" placeholder="' . esc_attr( wc_format_localized_price( 0 ) ) . '" class="refund_line_tax ceretax_line_total wc_input_price" size="4" data-line-tax="' . esc_attr( wc_format_localized_price( $total_line_tax ) ) . '" />
							</div>';
						}
					}
					echo '</td>';
				
				} elseif ( $item instanceof WC_Order_Item_Fee && $item->get_name() === $cere_tax_name ) { 
				// } elseif ( isset( $item['name'] ) && $item['name'] === $cere_tax_name ) {
					echo '<td class="ceretax_line_value" width="7%"></td>';
				} else {
					echo '<td class="ceretax_line_value" width="7%">';
						// echo '<div class="refund" style="display: none;">
						// 	<input type="text" name="ceretax_line_total[' . absint( $item_id ) . ']" placeholder="' . esc_attr( wc_format_localized_price( 0 ) ) . '" class="refund_line_tax ceretax_line_total wc_input_price" size="4" />
						// </div>';
					echo '</td>';
				}
			}
		}
	}

	add_action(
		'plugins_loaded',
		function () {
			cwafc()->admin->action = new CWAFC_Admin_Action();
		}
	);
}
